import z3
import z3_utils as z3u
import hcp_utils
import dfs
import stats


import spacer_args


def mk_pre_const (pred, arg): return _mk_arg_const (pred, arg, 0)
def mk_post_const (pred, arg): return _mk_arg_const (pred, arg, 1)

def _mk_arg_const (pred, arg, variant=0):
  """
  Returns a constant representing variant 'variant' argument 'arg' of
  predicate 'pred'
  """

  if not isinstance (pred, z3.FuncDeclRef) : pred = pred.decl ()

  name = "{}_{}_{}".format (pred.name (), arg, variant)
  return z3.Const (name, pred.domain (arg))

def _mk_const_variant (c, variant):
  return z3.Const (c.decl ().name () + '_' + str (variant),
                   c.sort ())
  
def _mk_pre_lemma (pred, lemma): return _mk_variant_lemma (pred, lemma, 0)
def _mk_post_lemma (pred, lemma): return _mk_variant_lemma (pred, lemma, 1)

@stats.time_stats
def _mk_variant_lemma (pred, lemma, variant=0):
  """
  Converts a lemma over a predicate to be over a given set of
  variables of that predicate
  """
  if pred is None: return lemma
  if not isinstance (pred, z3.FuncDeclRef) : pred = pred.decl ()

  args = [_mk_arg_const (pred, i, variant) for i in range (pred.arity ())]
  rw = z3u.SubstituteVars (*args)
  return rw (lemma)

def mk_post_lemma_variant (pred, lemma, variant):
  return _mk_variant_lemma (pred, lemma, variant=variant+1)

def ground_rule (rule_exp):
  """ Strip quantifier and replace bound vars by constants"""
  aux, matrix = z3u.stripQuantifierBlock (rule_exp)

  if z3u.isZ3Implies (matrix):
    body = matrix.arg (0)
    head = matrix.arg (1)
  else:
    head = matrix
    body = z3u.mk_true (rule_exp.ctx)

  return (aux, head, body)


def split_body (body):
  """ Splits body into Pred and Tail, where Pred is the only
  uninterpreted predicate (instance), and tail is all interpreted"""
  pred = z3u.getFirstConjunct (body)
  if pred.decl ().kind () != z3.Z3_OP_UNINTERPRETED:
    pred = None
    tail = body.children ()
  else:
    assert body.num_args () > 0
    tail = []
    if z3.is_and (body):
      tail.extend (body.children ()[1:])
  return pred, tail

def elim_pred (pred, variant=0) :
  if pred == None : return []
  pred_decl = pred.decl ()
  eqs = []
  for i in range (pred.num_args ()):
    const = _mk_arg_const (pred_decl, i, variant)
    arg = pred.arg (i)
    if not arg.eq (const):  eqs.append (arg == const)
  return eqs

def elim_pre_pred (pred): return elim_pred (pred, variant=0)
def elim_post_pred (pred): return elim_pred (pred, variant=1)

def elim_preds (head, pred, tail):
  res = []
  if pred != None:
    res = elim_pre_pred (pred)
    pred_decl = pred.decl ()
  else:
    pred_decl = None

  if head != None:
    res.extend (elim_post_pred (head))
    head_decl = head.decl ()
  else:
    head_decl = None
   
  return head_decl, pred_decl, res

def mk_pre_pred_consts (pdecl): return mk_pred_consts (pdecl, variant=0)
def mk_post_pred_consts (pdecl): return mk_pred_consts (pdecl, variant=1)

def mk_pred_consts (pdecl, variant=0) :
  return [_mk_arg_const (pdecl, i, variant) for i in range (pdecl.arity ())]

class Rule (z3u.HelperBase):
  def __init__ (self, rule_exp):
    z3u.HelperBase.__init__ (self, rule_exp.ctx)
    self.origin = rule_exp
    self.tag = None

    # parent pt 
    self.parent = None

    self.atrans = None
    self.amaker = None
    self.assump_manager = None
    self.used_assumptions = None

    self._init_from_rule_expr (rule_exp)
    self._exists_aux = z3u.MkExists (self.aux)
    self.qvars = None
    assert self.head.kind () == z3.Z3_OP_UNINTERPRETED
    assert self.src is None or \
        self.src.kind () == z3.Z3_OP_UNINTERPRETED

    
  def _init_from_rule_expr (self, rule_exp):
    """" Make first-order transition relation for the rule_exp """
    aux, head, body = ground_rule (rule_exp)
    pred, tail = split_body (body)

    hdecl, pdecl, state_eq =  elim_preds (head, pred, tail)

    # auxiliary variables
    self.aux = aux
    # the head predicate
    self.head = hdecl
    # the src predicate
    self.src = pdecl
    # the tail of the rule
    self.tail = tail
    # equalities connecting state-variables with auxiliary variables
    self.state_eq = state_eq

  def get_src_pt (self):
    if self.src is None: return None
    return self.get_problem ().get_pt (self.src)
  def get_head_pt (self):
    return self.parent

  def get_qvars (self):
    """vars to quantify"""
    if self.qvars is None:
      # include aux
      self.qvars = self.aux [:]
      # include pre consts for src
      if self.src is not None:
        src_pt = self.get_problem ().get_pt (self.src)
        self.qvars.extend (src_pt.pre_consts ())
      # include post consts for head
      self.qvars.extend (self.parent.post_consts ())
      self.qvars.extend (self.get_all_assumps ())
    return self.qvars

  def set_parent (self, p): self.parent = p
  def get_parent (self): return self.parent
  def get_problem (self): return self.parent.get_parent ()
  

  def get_src (self): return self.src
  def get_head (self): return self.head
  def get_origin (self): return self.origin

  def get_tag (self):
    if self.tag is not None : return self.tag
    if self.src is None : src_name = 'none'
    else : src_name = self.src.name ()
    self.tag = z3.Const ('{}_{}_tag'.format (src_name,
                                             self.head.name ()),
                         z3.BoolSort (ctx=self.ctx))
    return self.tag


  def get_assump_manager (self, amaker):
    if self.amaker is not None: assert self.amaker == amaker
    if self.amaker is None:
      self.amaker = amaker
      self.assump_manager = hcp_utils.AssumedLemmas (amaker, self.ctx)
    return self.assump_manager

  def is_concrete (self):
    return len (self.used_assumptions) == \
           len (self.assump_manager.get_assumps ())

  def concretize (self):
    self.used_assumptions = self.assump_manager.get_assumps ()

  def abstract_by_assumptions (self, assumps):
    ua = z3.AstVector (ctx=self.ctx)
    for a in assumps:
      if self.assump_manager.has_assump (a): 
        ua.push (a)
    self.used_assumptions = ua
        
  def refine_with_assumps (self, assumps):
    ua = self.used_assumptions
    am = self.get_assump_manager (self.get_problem ().mk_assume_var)
    for a in assumps:
      if am.has_assump (a) and a not in ua:
        #print 'REFINEMENT: ADDING:', a
        ua.push (a)

  #@stats.time_stats
  def mk_assumed_trans_variant(self, vnt):
    assert self.atrans is not None
    assert self.src is not None

    atrans = self.mk_and (*self.atrans)
    atrans = self._exists_aux (atrans)
    # copies for all aux vars
    #sub = [(v, _mk_const_variant (v, vnt)) for v in self.aux]
    sub = []
    src_pt = self.get_src_pt ()
    head_pt = self.get_head_pt ()
    # copies for all arg consts
    for i in range (self.src.arity ()):
      c = src_pt.pre_const (i)
      sub.append ( (c, _mk_arg_const (self.src, i, variant=vnt)) )
    for i in range (self.head.arity ()):
      c = head_pt.post_const (i)
      sub.append ( (c, _mk_arg_const (self.head, i, variant=vnt + 1)) )
    # substitute and return
    rw = z3u.Substitute (*sub)
    return [rw.apply (atrans)]

  def mk_assumed_trans (self, assump_maker):
    return self._mk_assumed_trans (assump_maker, self.tail + self.state_eq)

  def _mk_assumed_trans (self, assump_maker, trans, extra=None):
    """
    Creates a transition with assumptions. Assumptions are added for
    all elements of trans. extra can be used to provide additional
    constraints that should not be placed under any assumptions.
    """
    if self.atrans is None:
      assert self.assump_manager is None
      assert self.used_assumptions is None

      amanager = self.get_assump_manager (assump_maker)
      sa = hcp_utils.sprinkle_assumptions
      self.atrans = [sa (i, amanager) for i in trans]
      if extra is not None: self.atrans.extend (extra)
      self.used_assumptions = self.assump_manager.get_assumps ()

    return self.atrans, self.used_assumptions

  def get_all_assumps (self):
    return self.assump_manager.get_assumps ()

  def get_used_assumps (self):
    return self.used_assumptions
    

  def mk_rule (self):
    body = []
    if self.src is not None:
      src_pt = self.get_problem ().get_pt (self.src)
      body.append (src_pt.pre_inst ())
    trans, ua = self.mk_assumed_trans (self.get_problem ().mk_assume_var)
    body.extend (trans)
    body.extend (ua)

    matrix = self.parent.post_inst ()

    and_body = self.mk_and (*body)
    if not z3.is_true (and_body):
      matrix = z3.Implies (and_body, matrix)

    qvars = self.get_qvars ()

    return z3u.z3ForAll (qvars, matrix)


  def mk_rule_no_assumps (self):
    """
    similar to mk_rule above but resolving the assumptions away
    """
    body = []
    if self.src is not None:
      src_pt = self.get_problem ().get_pt (self.src)
      body.append (src_pt.pre_inst ())
    trans, ua = self.mk_assumed_trans (self.get_problem ().mk_assume_var)

    aa = self.assump_manager.get_assumps ()
    true_subs = [(a,z3.BoolVal (True, self.ctx)) for a in ua]
    false_subs = [(a,z3.BoolVal (False, self.ctx)) for a in aa]

    trans = trans[:]
    trans = [z3.substitute (t, *true_subs) for t in trans]
    trans = [z3.substitute (t, *false_subs) for t in trans]
        
    body.extend (trans)

    matrix = self.parent.post_inst ()

    and_body = self.mk_and (*body)
    and_body = z3.simplify (and_body)
    if not z3.is_true (and_body):
      matrix = z3.Implies (and_body, matrix)

    return z3u.z3ForAll (qvars, matrix)


  def box (self, *args, **kwrds): return self.mk_rule ()


  def __repr__ (self, no_assumps=False) :
      if no_assumps: return self.mk_rule_no_assumps ().sexpr ()
      else: return self.mk_rule ().sexpr ()


class BoundedRule (Rule):
  def __init__ (self, rule_exp, hb_decl, sb_decl) :
    assert isinstance (rule_exp, z3.BoolRef)
    self.head = hb_decl
    self.src = sb_decl
    Rule.__init__(self, rule_exp)
    
  def _init_from_rule_expr (self, rule_exp):
    self.aux, head, body = ground_rule (rule_exp)
    pred, tail = split_body (body)

    ### given H(a1, ..., a_n) and a decl HB(y_1, ..., y_m, x_1, x_n)
    ### create HB (c_1, ..., c_m, a_1, ..., a_n), where c_i is the
    ### post-constant for the ith argument of HB
    args = [mk_post_const (self.head, i) \
              for i in range (0, self.head.arity () - head.num_args ())]
    args.extend (head.children ())
    bhead = self.head (*args)

    ### same as above, but now for P and PB
    bsrc = None
    if self.src != None:
      args = [mk_pre_const (self.src, i) \
                for i in range (self.src.arity () - pred.num_args ())]
      args.extend (pred.children ())
      bsrc = self.src (*args)

    ### Eliminate HB() and PB() from tail. The first two return
    ### values are ignored
    __tmp1__, __tmp2__, self.state_eq = elim_preds (bhead, bsrc, tail)
    self.tail = tail


    # constraints for bounding
    self.bound_cons = []

  def get_bound_cons (self) : return self.bound_cons

  def add_bound_cons (self, cons) :
    """cons is a list of constraints"""
    self.bound_cons.extend (cons)

  @stats.time_stats
  def mk_assumed_trans_variant (self, vnt):
    atrans = Rule.mk_assumed_trans_variant (self, vnt)
    # create copy of bval_const
    bc = self.parent.bval_const
    if bc is not None:
      bc_vnt = _mk_const_variant (bc, vnt)
      rw = z3u.Substitute ((bc, bc_vnt))
      atrans = [rw.apply (conj) for conj in atrans]
    return atrans

  def mk_assumed_trans (self, assump_maker):

    abs_trans = []
    extra_trans = self.bound_cons[:]

    if spacer_args.args.no_trans_abs:
      extra_trans.extend (self.tail)
    else:
      abs_trans.extend (self.tail)

    if spacer_args.args.no_sig_abs:
      extra_trans.extend (self.state_eq)
    else:
      abs_trans.extend (self.state_eq)
    
    return self._mk_assumed_trans (assump_maker, 
                                   abs_trans,
                                   extra=extra_trans)

  def mk_rule (self, val=None, use_invs=False):
    """
    Create the rule and bound parent pt's counter to val.  If 'val' is
    None, it is not substituted but is quantified away
    """
    body = []
    if self.src is not None:
      src_pt = self.get_problem ().get_pt (self.src)
      body.append (src_pt.pre_inst ())
    trans, ua = self.mk_assumed_trans (self.get_problem ().mk_assume_var)

    body.extend (ua)
    del ua

    trans = trans[:]
    qvars = self.get_qvars ()[:]

    val_const = self.parent.bval_const
    if val_const is not None:
      if val is None:  
        qvars.append (val_const)

        #avar = self.get_problem ().mk_assume_var ()
        #qvars.append (avar)
        #trans.append (z3.Implies (avar, val_const == z3.IntVal (2, ctx=self.ctx)))
                                  
      else:
        # can be optimized by substituting for just 'the' constraint
        # which talks about the upper-bound constant
        #rw = z3u.Substitute ((val_const, val))
        #trans = [rw.apply (c) for c in trans]
     
        qvars.append (val_const)
        if not spacer_args.args.no_bounds:
          trans.append (val_const == val)
        
    body.extend (trans)
    if use_invs: body.extend (self.parent.mk_post_invs ())

    matrix = self.parent.post_inst ()

    and_body = self.mk_and (*body)
    if not z3.is_true (and_body):
      matrix = z3.Implies (and_body, matrix)

    return z3u.z3ForAll (qvars, matrix)

  def mk_rule_no_assumps (self, val=None, use_invs=False):
    """
    similar to mk_rule above but resolving the assumptions away
    """
    body = []
    if self.src is not None:
      src_pt = self.get_problem ().get_pt (self.src)
      body.append (src_pt.pre_inst ())
    trans, ua = self.mk_assumed_trans (self.get_problem ().mk_assume_var)

    trans = trans[:]
    qvars = self.get_qvars ()[:]

    val_const = self.parent.bval_const
    if val_const is not None:
      if val is None:  
        qvars.append (val_const)

        #avar = self.get_problem ().mk_assume_var ()
        #qvars.append (avar)
        #trans.append (z3.Implies (avar, val_const == z3.IntVal (2, ctx=self.ctx)))
                                  
      else:
        # can be optimized by substituting for just 'the' constraint
        # which talks about the upper-bound constant
        #rw = z3u.Substitute ((val_const, val))
        #trans = [rw.apply (c) for c in trans]
     
        qvars.append (val_const)
        if not spacer_args.args.no_bounds:
          trans.append (val_const == val)

    aa = self.assump_manager.get_assumps ()
    true_subs = [(a,z3.BoolVal (True, self.ctx)) for a in ua]
    false_subs = [(a,z3.BoolVal (False, self.ctx)) for a in aa]
    trans = [z3.substitute (t, *true_subs) for t in trans]
    trans = [z3.substitute (t, *false_subs) for t in trans]
        
    body.extend (trans)
    if use_invs: body.extend (self.parent.mk_post_invs ())

    matrix = self.parent.post_inst ()

    and_body = self.mk_and (*body)
    and_body = z3.simplify (and_body)
    if not z3.is_true (and_body):
      matrix = z3.Implies (and_body, matrix)

    return z3u.z3ForAll (qvars, matrix)

  def box (self, val=None, use_invs=False): 
    return self.mk_rule (val=val, use_invs=use_invs)

  def __repr__ (self, no_assumps=False) :
      if no_assumps: return self.mk_rule_no_assumps ().sexpr ()
      else: return self.mk_rule ().sexpr ()


class PredTransformer (z3u.HelperBase):
    def __init__ (self, pred):
        z3u.HelperBase.__init__ (self, pred.ctx)
        # unique id of the transformer 
        self.id = None
        # the head predicate (z3.FuncDeclRef)
        self.head = pred
        # rules defining the transformer
        self.rules = list ()

        ### AG: I don't think this is needed any more.
        ### AG: instead, store in each PT a pointer to the HC problem 
        ### AG: it is from (say under self.hcp). Then, to get the src for 
        ### AG: the rule do  self.hcp.get_pt (rule.srcPred ())
        ### AK: we can, but does it hurt? as we have users (below), we can have
        ###      this as well
        # src pts of the rules
        self.srcs = list ()

        # predicate transformers that use `self' in some body
        self.users = list ()

        # original transformer we came from
        self.origin = None

        # number of bound params in the signature of head
        self.num_bound_params = 0
        # bound Value
        self.bval_const = None

        self.tag = None
        self.pre_cond = None
        self.trans = None
        self.bg = None

        # invariants for pt
        self.invs = z3.AstVector (ctx=self.ctx)

        self.parent = None

        self._pre_consts = [_mk_arg_const (self.head, i, 0) 
                            for i in range (self.head.arity ())]
        self._post_consts = [_mk_arg_const (self.head, i, 1) 
                             for i in range (self.head.arity ())]
        self._pre_inst = self.head (*self._pre_consts)
        self._post_inst = self.head (*self._post_consts)

        self.pre_rewriter = z3u.SubstituteVars (*self._pre_consts)
        self.post_rewriter = z3u.SubstituteVars (*self._post_consts)

    def mk_pre_lemma (self, lemma):
      return self.pre_rewriter (lemma)
    def mk_post_lemma (self, lemma):
      return self.post_rewriter (lemma)

    def mk_pre_invs (self) :
      was_at_pt = self.get_pre_cond ()
      pre_invs = [z3.Implies (was_at_pt,
                              self.mk_pre_lemma (inv))
                  for inv in self.get_invs ()]
      return pre_invs

    def mk_pre_inv (self, inv):
      return z3.Implies (self.get_pre_cond (), self.mk_pre_lemma (inv))

    def mk_tagged_post_invs (self):
      post_invs = [z3.Implies (self.get_tag (),
                               self.mk_post_lemma (inv))
                   for inv in self.get_invs ()]
      return post_invs

    def mk_tagged_post_inv (self, inv):
      return z3.Implies (self.get_tag (), self.mk_post_lemma (inv))

    def mk_post_invs (self):
      post_invs = [self.mk_post_lemma (inv)
                   for inv in self.get_invs ()]
      return post_invs

    def pre_const (self, i):
      return self._pre_consts [i]
    def post_const (self, i):
      return self._post_consts [i]

    def pre_consts (self): return self._pre_consts
    def post_consts (self): return self._post_consts

    def pre_inst (self): return self._pre_inst
    def post_inst (self): return self._post_inst

    
    def populate_fp (self, fp, bval=None, use_invs=False, strengthen_trans=False):
      """ Adds rules to fp """
      ### box up the rules and add to fp
      assert self.ctx == fp.ctx

      bval_expr = None
      if bval is not None: bval_expr = z3.RealVal (bval, self.ctx)
      for r in self.rules:
        fp.add_rule (r.box (val=bval_expr, use_invs=strengthen_trans))

      ### add invs
      if use_invs:
        for inv in self.invs: fp.add_cover (-1, self.head, inv)
        if spacer_args.args.bound_lemmas and bval is not None and self.bval_const is not None:
          zero = z3.RealVal (0, self.ctx)
          cnt = z3.Var (self.bd_arg_pos, z3.RealSort (self.ctx))
          lemma1 = zero  <= cnt
          fp.add_cover (-1, self.head, lemma1)
          lemma2 = cnt <= bval_expr
          fp.add_cover (-1, self.head, lemma2)
          print 'BND LEMMAS', self.head, [lemma1, lemma2]
          

    def set_id (self, id) : self.id = id
    def get_id (self) : return self.id
    def set_parent (self, p): self.parent = p
    def get_parent (self): return self.parent

    def get_head (self): return self.head
    
    def add_user (self, pt):
        assert pt not in self.users
        self.users.append (pt)

    def add_src (self, pt) :
        assert pt not in self.srcs
        self.srcs.append (pt)

    def add_rule (self, rule) : 
      self.rules.append (rule)
      rule.set_parent (self)

    def get_rules (self): return self.rules
    def get_num_rules (self): return len (self.rules)

    def get_rule_from (self, pt) :
      if pt not in self.srcs : return None
      return self.rules [self.srcs.index (pt)]

    def user_rules (self):
      """generator for rules using self"""
      for user in self.users :
        r = user.get_rule_from (self)
        if r is not None : yield r

    def is_behind (self, u): return self.id <= u.id

    def create_bound_const (self, sort):
      assert self.bval_const is None
      self.bval_const = z3.Const ('{}!bval'.format (self.head.name ()), sort)

    def set_num_bound_params (self, num) : 
      self.num_bound_params = num
      # some large number, ideally larger than current bound
      max_int = 32767
      if spacer_args.args.up_counter:
        vals = [z3.RealVal (max_int, ctx=self.ctx) for i in range (num)]
      else:
        vals = [z3.RealVal (0, ctx=self.ctx) for i in range (num)]
      self.bound_rewriter = z3u.SubstituteVars (*vals)
      
    def get_num_bound_params (self): return self.num_bound_params

    def elim_bounds (self, lemma):
      return self.bound_rewriter (lemma)

    def set_origin_pt (self, pt) : self.origin = pt

    def get_tag (self):
      if self.tag is not None : return self.tag
      self.tag = z3.Const ('{}_tag'.format (self.head.name ()),
                           z3.BoolSort (ctx=self.ctx))
      return self.tag

    def get_pre_cond (self):
      """return constraint for being at self in the previous time step
      """
      if self.pre_cond is not None: return self.pre_cond
      out_tags = [r.get_tag () for r in self.user_rules ()]
      self.pre_cond = self.mk_or (*out_tags)
      return self.pre_cond

    def is_concrete (self):
      to_ret = True
      for r in self.rules:
        if not r.is_concrete ():
          to_ret = False
          break
      return to_ret

    def concretize (self):
      for r in self.rules: r.concretize ()

    def abstract_by_assumptions (self, assumps):
      for r in self.rules: 
        r.abstract_by_assumptions (assumps)
        print 'Rule of {} used {} assumptions out of {}'.format (self.head, 
                                                                 len (r.used_assumptions), 
                                                                 len (r.assump_manager.get_assumps ()))
        
    def mk_assumed_trans (self, assump_maker, get_assumps=True):
      assumps = None
      if self.trans is None:
        r_tags = [r.get_tag () for r in self.rules]
        self.trans = z3.Or (*r_tags)
  
        self.bg = []
        if get_assumps: assumps = z3.AstVector (ctx=self.ctx)
        for r in self.rules:
          r_trans, r_assumps = r.mk_assumed_trans (assump_maker)
          and_r_trans = self.mk_and (*r_trans)
          self.bg.append (z3.Implies (r.get_tag (), and_r_trans))
          if get_assumps: z3u.extend (assumps, r_assumps)
      elif get_assumps:
        assumps = self.get_used_assumps ()

      return (self.trans, self.bg, assumps)
    
    def get_all_assumps (self):
      assumps = z3.AstVector (ctx=self.ctx)
      for r in self.rules:
        z3u.extend (assumps, r.get_all_assumps ())
      return assumps

    def get_used_assumps (self):
      ua = z3.AstVector (ctx=self.ctx)
      for r in self.rules:
        z3u.extend (ua, r.get_used_assumps ())
      return ua

    def add_inv (self, inv, add_to_solver=True):
      if inv not in self.invs:
        self.invs.push (inv)
        if add_to_solver:
          self.parent.add_to_solver (self.mk_pre_inv (inv))
          self.parent.add_to_solver (self.mk_tagged_post_inv (inv))

    def is_inv (self, lem):
      return (lem in self.invs)

    def get_invs (self) : return self.invs

    def __repr__ (self): return repr (self.head)
    def __repr_verbose__ (self) :
        s = ''
        for rule in self.rules :
            s = s + rule.__repr__ (no_assumps=
                                   spacer_args.args.print_no_assumps) \
                  + '\n'
        return s
    
    # graph.Node interface
    def get_succs (self): return iter (self.users)
    def has_succ (self, n): return n in self.users


class HornClauseProblem (object):
    def __init__ (self, ctx=None):
      
      self.ctx = ctx
      if self.ctx == None: self.ctx = z3.main_ctx ()
      
      # predicate transformers
      self.pts = list ()
      # map from preds to pts
      self.ptMap = dict ()
      # special predicate transformers
      self.root = None
      self.query = None
      # true when closed
      self.closed = False
      # heterogeneous list of lists and pts
      self.wto = None

      self.num_assume_vars = 0

      self.trans = None

      self.solver = None

    def _createWto (self) :
      def visit (pt, curr_list, id) :
        pt.set_id (id)
        curr_list.append (pt)
        return {'id' : id + 1}

      def backedge (from_pt, pt, wto, curr_list) :
        l = wto
        while (pt not in l) : l = l[-1]
        idx = l.index (pt)
        # if pt is already a head, do nothing
        if idx == 0 : return None
        # the suffix from idx forms a new list
        new_list = l [idx:]
        # replace the suffix from idx with the new list
        l [idx:] = list ()
        l.append (new_list)
        if curr_list == l :
          curr_list = new_list
        return {'curr_list' : curr_list}

      # empty list
      self.wto = list ()

      traversal = dfs.DFS (lambda n : z3u.z3AstRefKey (n.head), lambda n : n.users)
      traversal.addHandler (visit, ['curr_list', 'id'], 'visit')
      traversal.addHandler (backedge, ['wto', 'curr_list'], 'backedge')
      traversal.setInitValue ('curr_list', self.wto)
      traversal.setInitValue ('wto', self.wto)
      traversal.setInitValue ('id', 0)
      traversal.search (self.root)

      # move query to the end of wto
      for wto_comps, pt in hcp_utils.wtoWalk (self.wto) :
        if pt == self.query :
          assert pt in wto_comps [-1]
          wto_comps [-1].remove (pt)
          break
      self.wto.append (self.query)

    def close (self):
        if self.closed : return
        assert self.root is not None
        assert self.query is not None

        # create wto numbering all pts
        if self.wto is None :
          self._createWto ()

        # mark the problem closed
        self.closed = True

    def add_pt (self, pt):
        self.pts.append (pt)
        self.ptMap [z3u.z3AstRefKey (pt.head)] = pt
        pt.set_parent (self)

    def add_rule (self, rule):
      """ adds a rule """
      ## add rule to PT that owns it
      head_pt = self.get_pt (rule.get_head ())
      assert head_pt is not None
      head_pt.add_rule (rule)

      ## update srcPT if it exists
      src_pred = rule.get_src ()
      if src_pred is None : src_pt = None
      else : src_pt = self.get_pt (src_pred)
      if src_pt is not None :
        head_pt.add_src (src_pt)
        src_pt.add_user (head_pt)
      else: head_pt.add_src (None)

    def ensure_pt (self, pred) :
        key = z3u.z3AstRefKey (pred)
        if key in self.ptMap : return self.ptMap [key]
        pt = PredTransformer (pred)
        self.add_pt (pt)
        return pt

    def get_pt (self, pred) :
        key = z3u.z3AstRefKey (pred)
        return self.ptMap.get (key)

    def set_root (self, pt):
        assert self.root is None
        self.root = pt

    def get_root (self): return self.root
    def get_query (self): return self.query

    def get_query_pred (self): return self.query.head ()

    def set_query (self, pt):
        assert self.query is None
        self.query = pt

    def is_concrete (self):
      to_ret = True
      for pt in self.pts:
        if not pt.is_concrete ():
          to_ret = False
          break
      return to_ret

    def concretize (self):
      for pt in self.pts: pt.concretize ()

    def abstract_by_assumptions (self, assumps):
      for pt in self.pts: pt.abstract_by_assumptions (assumps)

    def get_all_assumps (self):
      assumps = z3.AstVector (ctx=self.ctx)
      for pt in self.pts:
        z3u.extend (assumps, pt.get_all_assumps ())
      return assumps

    def get_used_assumps (self):
      ua = z3.AstVector (ctx=self.ctx)
      for pt in self.pts:
        z3u.extend (ua, pt.get_used_assumps ())
      return ua

    def mk_assumed_trans (self, get_assumps=True):
      assumps = None
      if self.trans is None:
        self.trans = []
        if get_assumps: assumps = z3.AstVector (ctx=self.ctx)
        for pt in self.pts :
          pt_trans, pt_bg, pt_assumps = \
             pt.mk_assumed_trans (self.mk_assume_var, get_assumps=get_assumps)
          self.trans.extend (pt_bg)
          self.trans.append (z3.Implies (pt.get_tag (), pt_trans))
          if get_assumps: z3u.extend (assumps, pt_assumps)
      elif get_assumps:
        assumps = self.get_used_assumps ()
      return self.trans, assumps
      
    def mk_assume_var (self):
      v = z3.Const ('A!{:x}!{}'.format (id (self), self.num_assume_vars), z3.BoolSort (ctx=self.ctx))
      self.num_assume_vars += 1
      return v

    def get_solver (self, get_assumps=True):
      assumps = None
      if self.solver is None :
        self.solver = z3.Solver (ctx=self.ctx)
        ### problem skeleton
        trans, assumps = self.mk_assumed_trans (get_assumps=get_assumps)
        self.solver.add (*trans)
      elif get_assumps:
        assumps = self.get_used_assumps ()
      return self.solver, assumps

    def add_to_solver (self, exp):
      assert self.solver is not None
      self.solver.add (exp)

    def __repr__ (self) :
        s = 'Root : {}\n'.format (repr (self.root.head))
        for pt in self.pts : s = s + pt.__repr_verbose__ ()
        s = s + 'Query : {}\n'.format (repr (self.query.head))
        return s

    # graph.DirectedGraph interface
    def get_nodes (self): return iter (self.pts)




@stats.time_stats
def create_hc_problem (filename, pp=False, z3ctx=None, dummy_query=False) :
  fp = z3.Fixedpoint (ctx=z3ctx)
  if not pp:
    print 'No pre-processing'
    fp.set (slice=False)
    fp.set (inline_linear=False)
    fp.set (inline_eager=False)

  with stats.timer ('fp.parse_file'):
    q = fp.parse_file (filename)
  
  if dummy_query:
    with stats.timer ('fp.dummy_query'):
      fp.query (z3u.mk_true (z3ctx))


  prob = HornClauseProblem (z3ctx)

  rules = list ()

  # create pts
  for rule_exp in fp.get_rules () :
    rule = Rule (rule_exp)
    head_pred = rule.get_head ()
    pt = prob.ensure_pt (head_pred)
    if rule.get_src () is None: prob.set_root (pt)
    rules.append (rule)

  # add rules to prob
  for rule in rules : prob.add_rule (rule)

  # only one query for now
  assert len(q) == 1
  pt = prob.get_pt (q[0].decl ())
  assert pt is not None
  prob.set_query (pt)


  del fp
  return prob

if __name__ == '__main__':
  import sys
  create_hc_problem (sys.argv[1])
